--[[

Interface: 1.5.1.0 b1580

Author: GtX | Andy (LS-MODCOMPANY)

http://ls-modcompany.com

Special Thanks: Giants Software (Some parts based on original Cover.lua)

Date: 01.08.2017

Version: 1.0.0.0 - Initial Version.
Version: 1.0.1.0 - Added clock for digital display.
Version: 1.0.2.0 - Added pipe overwrite function to allow the `Kotte FRC' to be filled from the pipe with effects.

About: Designed to operate cover on tankers. Extends control of overload pipe and cover to outside of trailer. Also allows external only operation of cover. Adds extra design registration as needed by mod.
	   Also adds a clock option for operation screens.

	   ** Free for use on other mods **

]]--

TankTech = {}

function TankTech.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(AnimatedVehicle, specializations)
end

function TankTech:preLoad(savegame)
    self.getIsToggleCoverAllowed = TankTech.getIsToggleCoverAllowed
    self.allowFillType = Utils.overwrittenFunction(self.allowFillType, TankTech.allowFillType)
    self.getAllowFillFromAir = Utils.overwrittenFunction(self.getAllowFillFromAir, TankTech.getAllowFillFromAir)
    self.setCoverState = TankTech.setCoverState
	self.isInTriggerNode = TankTech.isInTriggerNode	
end

function TankTech:load(savegame)
	
	if self.configurations["tankWrap"] ~= nil then
		ObjectChangeUtil.updateObjectChanges(self.xmlFile, "vehicle.tankWrapConfigurations.tankWrapConfiguration", self.configurations["tankWrap"], self.components, self)
    end

	if self.configurations["fuelBrand"] ~= nil then
		ObjectChangeUtil.updateObjectChanges(self.xmlFile, "vehicle.fuelBrandConfigurations.fuelBrandConfiguration", self.configurations["fuelBrand"], self.components, self)
    end
	
	local index = "vehicle.tankTech"
	
	-- Overwrite Function (findTrailerRaycastCallback) to allow the Base Game `Kotte FRC` to be loaded with effect using the Pipe.lua
	local frcOverwriteFunction = Utils.getNoNil(getXMLBool(self.xmlFile, index..".pipe#frcOverwriteFunction"), false)
	if frcOverwriteFunction then
		self.findTrailerRaycastCallback = Utils.overwrittenFunction(self.findTrailerRaycastCallback, TankTech.findTrailerRaycastCallback)
		-- Extend the z axis of effect when overloading to the Kotte FRC ( BASE GAME ONLY ).
		self.frcExtension = Utils.getNoNil(getXMLFloat(self.xmlFile, index..".pipe#frcEffectExtension"), 5)
	end

	self.coverAnimation = getXMLString(self.xmlFile, index..".cover#animationName")

    if self.coverAnimation ~= nil then
        local isOpen = Utils.getNoNil(getXMLBool(self.xmlFile, index..".cover#openOnBuy"), false)
        if savegame ~= nil then
            isOpen = Utils.getNoNil(getXMLBool(savegame.xmlFile, savegame.key.."#isCoverOpen"), isOpen)
        end
        self:setCoverState(isOpen, true)
        self.autoReactToTrigger = Utils.getNoNil(getXMLBool(self.xmlFile, index..".cover#autoReactToTrigger"), false)
		self.insideCoverControl = Utils.getNoNil(getXMLBool(self.xmlFile, index..".cover#insideCoverControl"), true)
    else
        self.isCoverOpen = true
    end

	self.tt = {}
	self.tt.index = Utils.indexToObject(self.components, getXMLString(self.xmlFile, index..".externalControl#triggerIndex"))
	self.tt.triggerDistance = Utils.getNoNil(getXMLFloat(self.xmlFile, index..".externalControl#triggerDistance"), 1.0)

	self.tt.externalCoverActive = Utils.getNoNil(getXMLBool(self.xmlFile, index..".externalCover#active"), false)
	self.tt.externalPipeActive = Utils.getNoNil(getXMLBool(self.xmlFile, index..".externalPipe#active"), false)
	self.tt.externalOverActive = Utils.getNoNil(getXMLBool(self.xmlFile, index..".externalOverload#active"), false)

	if self.tt.index ~= nil then
		self.tt.coverOpenText = Utils.getNoNil(getXMLString(self.xmlFile, index..".externalCover#openText"), "action_openTankerCover")
		self.tt.coverCloseText = Utils.getNoNil(getXMLString(self.xmlFile, index..".externalCover#closeText"), "action_closeTankerCover")

		self.tt.pipeOut = Utils.getNoNil(getXMLString(self.xmlFile, index..".externalPipe#unfoldText"), "action_pipeOut")
		self.tt.pipeIn = Utils.getNoNil(getXMLString(self.xmlFile, index..".externalPipe#foldText"), "action_pipeIn")

		self.tt.startOverload = Utils.getNoNil(getXMLString(self.xmlFile, index..".externalOverload#startText"), "action_startOverloading")
		self.tt.stopOverload = Utils.getNoNil(getXMLString(self.xmlFile, index..".externalOverload#stopText"), "action_stopOverloading")
	end

	local hudDisplay = "vehicle.tankTech.hudDisplay"

	self.tt.timeHud = VehicleHudUtils.loadHud(self, self.xmlFile, "clock", hudDisplay)
	if self.tt.timeHud ~= nil then
        self.minuteChanged = Utils.appendedFunction(self.minuteChanged, TankTech.minuteChanged)
        g_currentMission.environment:addMinuteChangeListener(self)
        self:minuteChanged()
    end
end

function TankTech:getSaveAttributesAndNodes(nodeIdent)
    if self.coverAnimation ~= nil then
        local attributes = 'isCoverOpen="'..tostring(self.isCoverOpen)..'"'
        return attributes, nil
    end
    return nil
end

function TankTech:readStream(streamId, connection)
    if connection:getIsServer() then
        if self.coverAnimation ~= nil then
            local isOpen = streamReadBool(streamId)
            self:setCoverState(isOpen, true)
        end
    end
end

function TankTech:writeStream(streamId, connection)
    if not connection:getIsServer() then
        if self.coverAnimation ~= nil then
            streamWriteBool(streamId, self.isCoverOpen)
        end
    end
end

function TankTech:delete()
	if self.tt.timeHud ~= nil then
        g_currentMission.environment:removeMinuteChangeListener(self)
    end
end

function TankTech:update(dt)
	if self.tt.index ~= nil then
		if self:isInTriggerNode() then

			if self.coverAnimation ~= nil and self.tt.externalCoverActive == true then
				if self.isCoverOpen then
					g_currentMission:addHelpButtonText(g_i18n:getText(self.tt.coverCloseText), InputBinding.TOGGLE_COVER, nil, GS_PRIO_HIGH)
				else
					g_currentMission:addHelpButtonText(g_i18n:getText(self.tt.coverOpenText), InputBinding.TOGGLE_COVER, nil, GS_PRIO_HIGH)
				end

				if InputBinding.hasEvent(InputBinding.TOGGLE_COVER) then
					if self:getIsToggleCoverAllowed(not self.isCoverOpen) then
						self:setCoverState(not self.isCoverOpen)
					end
				end
			end

			if self.hasPipe and self.tt.externalPipeActive == true then
				if self.pipeNumStates == 2 then
					if self.pipeTargetState == 2 then
						if self:getIsPipeStateChangeAllowed(1) then
							g_currentMission:addHelpButtonText(g_i18n:getText(self.tt.pipeIn), InputBinding.TOGGLE_PIPE, nil, GS_PRIO_HIGH)
						end
					else
						if self:getIsPipeStateChangeAllowed(2) then
							g_currentMission:addHelpButtonText(g_i18n:getText(self.tt.pipeOut), InputBinding.TOGGLE_PIPE, nil, GS_PRIO_HIGH)
						end
					end
				end

				if InputBinding.hasEvent(InputBinding.TOGGLE_PIPE) then
					local nextState = self.pipeTargetState+1
					if nextState > self.pipeNumStates then
						nextState = 1
					end
					if self:getIsPipeStateChangeAllowed(nextState) then
						self:setPipeState(nextState)
					elseif nextState ~= 1 and self:getIsPipeStateChangeAllowed(1) then
						self:setPipeState(1)
					end
				end
			end

			if self.overloading ~= nil and self.tt.externalOverActive == true then
				if self.overloading.canToggleOverloading and self:getIsOverloadingAllowed() then
					if self.overloading.isActive then
						g_currentMission:addHelpButtonText(g_i18n:getText(self.tt.stopOverload), InputBinding.TOGGLE_TIPSTATE, nil, GS_PRIO_HIGH)
					else
						if self:getUnitFillLevel(self.overloading.fillUnitIndex) > 0 and self.targetTrailer ~= nil then
							g_currentMission:addHelpButtonText(g_i18n:getText(self.tt.startOverload), InputBinding.TOGGLE_TIPSTATE, nil, GS_PRIO_HIGH)
						end
					end
				end

				if InputBinding.hasEvent(InputBinding.TOGGLE_TIPSTATE) and self.overloading.canToggleOverloading then
					if self:getIsOverloadingAllowed() then
						self:setOverloadingActive(not self.overloading.isActive)
					end
				end
			end
		end
	end

	if self.insideCoverControl then
		if self:getIsActiveForInput() then
			if self.coverAnimation ~= nil then
				if InputBinding.hasEvent(InputBinding.TOGGLE_COVER) then
					if self:getIsToggleCoverAllowed(not self.isCoverOpen) then
						self:setCoverState(not self.isCoverOpen)
					end
				end
			end
        end
	end
end

function TankTech:draw()
	if self.insideCoverControl then
		if self.isClient and self:getIsActiveForInput(true) then
			if self.coverAnimation ~= nil and self:getIsToggleCoverAllowed(not self.isCoverOpen) then
				if self.isCoverOpen then
					g_currentMission:addHelpButtonText(g_i18n:getText(self.tt.coverCloseText), InputBinding.TOGGLE_COVER, nil, GS_PRIO_NORMAL)
				else
					g_currentMission:addHelpButtonText(g_i18n:getText(self.tt.coverOpenText), InputBinding.TOGGLE_COVER, nil, GS_PRIO_NORMAL)
				end
			end
		end
	end
end

function TankTech:getIsToggleCoverAllowed(isCoverOpen)
    if self.tipState ~= nil then
        return self.tipState == Trailer.TIPSTATE_CLOSED
    end
    return true
end

function TankTech:allowFillType(superFunc, fillType, allowEmptying)
    if not self.isCoverOpen then
        return false
    end
    if superFunc ~= nil then
        return superFunc(self, fillType, allowEmptying)
    end
    return true
end

function TankTech:setCoverState(isOpen, noEventSend)
    if self.coverAnimation ~= nil then
        TankTechStateEvent.sendEvent(self, isOpen, noEventSend)
        self.isCoverOpen = isOpen
        if isOpen then
            self:playAnimation(self.coverAnimation, -1, self:getAnimationTime(self.coverAnimation), true)
        else
            self:playAnimation(self.coverAnimation, 1, self:getAnimationTime(self.coverAnimation), true)
        end
    end
end

function TankTech:mouseEvent(posX, posY, isDown, isUp, button)
end

function TankTech:keyEvent(unicode, sym, modifier, isDown)
end

function TankTech:getAllowFillFromAir(superFunc)
    if not self.isCoverOpen then
        return false
    end
    if superFunc ~= nil then
        return superFunc(self)
    end
    return true
end

function TankTech:isInTriggerNode()
	if self.tt.index ~= nil then
		local wx, wy, wz = getWorldTranslation(g_currentMission.player.rootNode)
		local tx, ty, tz = getWorldTranslation(self.tt.index)
		local distance = Utils.vector3Length(wx-tx, wy-ty, wz-tz)

		if distance <= self.tt.triggerDistance then
			return true
		else
			return false
		end
	end
end

if ConfigurationUtil.configurations["tankWrap"] == nil then
	ConfigurationUtil.registerConfigurationType('tankWrap', g_i18n:getText('configuration_tankWrap'), nil, Vehicle.getConfigColorSingleItemLoad, Vehicle.getConfigColorPostLoad, ConfigurationUtil.SELECTOR_MULTIOPTION)
end

if ConfigurationUtil.configurations["fuelBrand"] == nil then
	ConfigurationUtil.registerConfigurationType('fuelBrand', g_i18n:getText('configuration_fuelBrand'), nil, Vehicle.getConfigColorSingleItemLoad, Vehicle.getConfigColorPostLoad, ConfigurationUtil.SELECTOR_MULTIOPTION)
end

function TankTech:minuteChanged()
    local minutes = g_currentMission.environment.currentMinute
    local hours = g_currentMission.environment.currentHour
    local minutesString = string.format("%02d", minutes)
    VehicleHudUtils.setHudValue(self, self.tt.timeHud, tonumber(hours.."."..minutesString), 9999)
end

function TankTech:findTrailerRaycastCallback(superFunc, transformId, x, y, z, distance)
    local vehicle = g_currentMission.nodeToVehicle[transformId]
	self.pipeParticleSystemExtraDistanceOnTrailer = 0
	
    if vehicle ~= nil and vehicle ~= self then
        if self.trailerFound ~= 0 and self.trailerFoundDistance ~= 0 then
            return false
        end

        self.pipeFoundTrailer = true

		local storeItem = StoreItemsUtil.storeItemsByXMLFilename[vehicle.configFileName:lower()]
		if storeItem ~= nil and storeItem.name == "FRC" and not storeItem.isMod then
			self.pipeParticleSystemExtraDistanceOnTrailer = self.pipeParticleSystemExtraDistanceOnTrailer + self.frcExtension
			self.trailerFound = transformId
			self.trailerFoundDistance = distance+self.pipeParticleSystemExtraDistanceOnTrailer
		else
			if vehicle.exactFillRootNode == transformId then
				local fillType = self:getUnitLastValidFillType(self.overloading.fillUnitIndex)
				if vehicle:allowFillType(fillType) and vehicle.exactFillRootNode ~= nil then
					if vehicle:getFillLevel(fillType) < vehicle:getCapacity(fillType) then
						self.trailerFound = transformId
					end
				end
			else
				self.trailerFoundDistance = distance+self.pipeParticleSystemExtraDistanceOnTrailer
			end
		end
    else
        local object = g_currentMission:getNodeObject(transformId)
        if object ~= nil and object.parent ~= nil and object.parent.triggerId ~= nil then
            if object.parent:getAllowFillTypeFromTool(self:getUnitLastValidFillType(self.overloading.fillUnitIndex), TipTrigger.TOOL_TYPE_PIPE) then
                self.shovelTargetFound = object
                self.trailerFoundDistance = distance+self.pipeParticleSystemExtraDistanceOnTrailer
            end
        end
    end
	
	if superFunc ~= nil then
        return superFunc(self, transformId, x, y, z, distance)
    end
	
    return true
end

-- EVENT

TankTechStateEvent = {}
TankTechStateEvent_mt = Class(TankTechStateEvent, Event)
InitEventClass(TankTechStateEvent, "TankTechStateEvent")

function TankTechStateEvent:emptyNew()
    local self = Event:new(TankTechStateEvent_mt)
    return self
end

function TankTechStateEvent:new(object, isOpen)
    local self = TankTechStateEvent:emptyNew()
    self.object = object
    self.isOpen = isOpen
    return self
end

function TankTechStateEvent:readStream(streamId, connection)
    self.object = readNetworkNodeObject(streamId)
    self.isOpen = streamReadBool(streamId)
    self:run(connection)
end

function TankTechStateEvent:writeStream(streamId, connection)
    writeNetworkNodeObject(streamId, self.object)
    streamWriteBool(streamId, self.isOpen)
end

function TankTechStateEvent:run(connection)
    if not connection:getIsServer() then
        g_server:broadcastEvent(self, false, connection, self.object)
    end
    if self.object ~= nil then
        self.object:setCoverState(self.isOpen, true)
    end
end

function TankTechStateEvent.sendEvent(vehicle, isOpen, noEventSend)
    if vehicle.isCoverOpen ~= isOpen then
        if noEventSend == nil or noEventSend == false then
            if g_server ~= nil then
                g_server:broadcastEvent(TankTechStateEvent:new(vehicle, isOpen), nil, nil, vehicle)
            else
                g_client:getServerConnection():sendEvent(TankTechStateEvent:new(vehicle, isOpen))
            end
        end
    end
end